/*-
 * Copyright (C) 2006-2014 n13i <n13i@m2hq.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/**
 * Chararina plug-in skeleton
 *
 * $Id: plugin.cpp 1 2007-11-04 17:38:04Z naoh $
 */

#include "stdafx.h"

//#include <windows.h>
//#include <stdio.h>
#include <malloc.h>

#include "plugin.h"
#include "mecabdll.h"
#include "resource.h"

#define snprintf _snprintf_s


/* vOCCX^X̎f[^ */
typedef struct {
	char  pluginPath[MAX_PATH];
	char  datPath[MAX_PATH];
	char  rcFile[MAX_PATH + 9];
	char  userDic[MAX_PATH + 10];
	BOOL  useUserDic;
	char *lastError;
} PData;


#ifdef __cplusplus
extern "C"
{
#endif

/* vOC DLL ̃CX^X */
HINSTANCE hInstance;

/* {̂̃VXe֐̍\ */
struct TPWPFunctions expf;

/* libmecab.dll ̊֐̍\ */
struct TMeCabDLLFunctions *mf = NULL;

static int setLastError(PData *pdata, mecab_t *m);


/*
 * --------------------------------------------------------------------------
 * D֐
 * --------------------------------------------------------------------------
 */
/**
 * import int NH_MA_useUserDictionary(string)
 */
void PWPCALL
NH_MA_useUserDictionary(TPWPInstance *instance, const char *pcArgType,
	IRTValue **pArgs, int nArg, IRTValue *pRetVal, DWORD param)
{
	PData *pdata = (PData *)(instance->pdata);

	const char *path = pArgs[0]->getString();

	if(path == NULL)
	{
		pdata->userDic[0] = '\0';
		pdata->useUserDic = FALSE;
		pRetVal->setInt(1);
		return;
	}

	snprintf(pdata->userDic, sizeof(pdata->userDic),
			 "--userdic=%s\\%s", pdata->datPath, path);
//	MessageBox(NULL, pdata->userDic, PLUGIN_NAME, MB_OK);
	pdata->useUserDic = TRUE;
	pRetVal->setInt(1);
	return;
}

/**
 * import int NH_MA_parse(string, string[][])
 */
void PWPCALL
NH_MA_parse(TPWPInstance *instance, const char *pcArgType,
	IRTValue **pArgs, int nArg, IRTValue *pRetVal, DWORD param)
{
	PData *pdata = (PData *)(instance->pdata);

	int i;
	mecab_t *m;
	mecab_node_t *node;

	char *args[] = {
		"",
		pdata->rcFile,
		pdata->userDic,
	};


	const char *input = pArgs[0]->getString();
	IRTArray *ar = pArgs[1]->getArray();

	if(input == NULL)
	{
		pRetVal->setInt(0);
		return;
	}

	if((m = mf->mecab_new((pdata->useUserDic == FALSE ? 2 : 3), args)) == NULL)
	{
		setLastError(pdata, m);

		pRetVal->setInt(0);
		return;
	}

	ar->emptyof();

	if((node = mf->mecab_sparse_tonode(m, input)) == NULL)
	{
		setLastError(pdata, m);
		mf->mecab_destroy(m);
		pRetVal->setInt(0);
		return;
	}

	for(i = 0; node; node = node->next, i++)
	{
		char surface[1025]; // FIXME

		if(node->stat == MECAB_BOS_NODE || node->stat == MECAB_EOS_NODE)
		{
			surface[0] = '\0';
		}
		else
		{
			if(node->length >= 1024)
			{
				mf->mecab_destroy(m);
				pRetVal->setInt(0);
				return;
			}

			strncpy_s(surface, 1025, node->surface, node->length);
			surface[node->length] = '\0';
		}

		ar->getArrayAt(i)->setStringAt(0, surface);
		ar->getArrayAt(i)->setStringAt(1, node->feature);
	}

	mf->mecab_destroy(m);

	pRetVal->setInt(1);
}


/**
 * import string NH_MA_getLastError()
 */
void PWPCALL
NH_MA_getLastError(TPWPInstance *instance, const char *pcArgType,
	IRTValue **pArgs, int nArg, IRTValue *pRetVal, DWORD param)
{
	pRetVal->setString(((PData *)(instance->pdata))->lastError);
}


static int setLastError(PData *pdata, mecab_t *m)
{
	const char *error = mf->mecab_strerror(m);

	if((pdata->lastError = (char *)malloc(strlen(error) + 1)) == NULL)
	{
		return 0;
	}

//	MessageBox(NULL, error, PLUGIN_NAME, MB_OK);

	strncpy_s(pdata->lastError, strlen(error) + 1, error, strlen(error));
	return 1;
}



/*
 * --------------------------------------------------------------------------
 * j[R[obN
 * --------------------------------------------------------------------------
 */
BOOL CALLBACK
DialogProc(HWND hDlgWnd, UINT msg, WPARAM wp, LPARAM lp)
{
	TCHAR buf[1024];

	switch(msg)
	{
	case WM_CLOSE:
		EndDialog(hDlgWnd, IDCANCEL);
		return TRUE;
	case WM_INITDIALOG:
		wsprintf(buf, "Version %s", PLUGIN_VERSION_STRING);
		SetWindowText(GetDlgItem(hDlgWnd, IDC_PLUGINVER), buf);
		wsprintf(buf, "Powered by MeCab %s", mf->mecab_version());
		SetWindowText(GetDlgItem(hDlgWnd, IDC_MECABVER), buf);
		return TRUE;
	case WM_COMMAND:
		if(LOWORD(wp) == IDOK) EndDialog(hDlgWnd, IDOK);
		else EndDialog(hDlgWnd, IDCANCEL);
		return TRUE;
	}
	return FALSE;
}

void PWPCALL
OnConfClick(TPWPInstance *instance, const char *item)
{
//	char buf[1024];

	DialogBox(hInstance, TEXT("ABOUTDLG"),
			  expf.PWPGetWindowHandle(instance, WHT_APPLICATION),
			  (DLGPROC)DialogProc);

/*
	snprintf(buf, sizeof(buf),
			 "`ԑf̓vOC Version %d.%d\n"
			 "Copyright (C) 2006 mxproject\n\n"
			 "libmecab-1.dll: Version %s",
			 PLUGIN_VERSION_MAJOR, PLUGIN_VERSION_MINOR,
			 ((PData *)(instance->pdata))->mf->mecab_version());

	MessageBox(NULL,
			   buf,
			   PLUGIN_NAME, MB_OK);
*/
}


/*
 * --------------------------------------------------------------------------
 * K{R[obN
 * --------------------------------------------------------------------------
 */

/**
 * vOC̏
 */
PWPEXPORT BOOL PWPCALL
PWCInitialize(int major, int minor, const char *substr,
			  const struct TPWPFunctions *p)
{
	if(major < 3)
	{
		MessageBox(NULL,
			TEXT("̃o[W Chararina ł͗pł܂B"),
			PLUGIN_NAME, MB_OK);
		return FALSE;
	}

	/*
	 * VXe֐\̂̃Rs[
	 */
	unsigned int sz = sizeof(struct TPWPFunctions);
	memset(&expf, 0, sz);

	if (p->size < sz) { sz = p->size;  }

	memcpy(&expf, p, sz);

	return TRUE;
}

/**
 * vOCCX^X
 * 1 LN^ 1 
 */
PWPEXPORT int PWPCALL
PWCNewInstance(TPWPInstance *instance, const char *vendor, const char *type)
{
	PData *pdata;
	char buf[1024];

	instance->pdata = NULL;

	/*
	 * instance->pdata ɕKvȃf[^Zbg
	 */
	if((pdata = (PData *)malloc(sizeof(PData))) == NULL)
	{
		MessageBox(NULL,
			TEXT("vOCCX^X̍쐬Ɏs܂B"),
			PLUGIN_NAME, MB_OK);

		return 0;
	}


	/* vOCtH_𓾂 */
	expf.PWPGetPluginFolder(instance, pdata->pluginPath);

	snprintf(pdata->rcFile, sizeof(pdata->rcFile),
			 "--rcfile=%s\\mecab\\mecabrc", pdata->pluginPath);
//	snprintf(pdata->sysDic, sizeof(pdata->sysDic),
//			 "--dicdir=%s\\dic\\ipadic", pdata->pluginPath);
	pdata->userDic[0] = '\0';
	pdata->useUserDic = 0;
	pdata->lastError = NULL;

	/* LN^ dat tH_̃pXݒ */
	expf.PWPGetPersonaDataFolder(instance, pdata->datPath);
//	MessageBox(NULL, pdata->datPath, PLUGIN_NAME, MB_OK);


	/*
	 * MeCab [hĂȂ΂Ń[hE
	 */
	if(!mf)
	{
		snprintf(buf, sizeof(buf),
				 "%s\\mecab\\libmecab.dll", pdata->pluginPath);

		if((mf = MeCabDLL_load(buf)) == NULL)
		{
			MessageBox(NULL,
					   TEXT("libmecab.dll ̃[hɎs܂B"),
					   "", MB_OK);

			free(instance->pdata);
			return 0;
		}
	}


	/*
	 * j[o^
	 */
	snprintf(buf, sizeof(buf), "%s\\`ԑf̓vOCɂ",
			 expf.PWPMenuItem_GetSetupMenuName(instance));

	expf.PWPMenuItem_Add(instance, buf, MS_NORMAL, 0, OnConfClick);


	/*
	 * ֐o^
	 */
	expf.PWPRegisterAyaoriFunc2(
		instance,
		"NH_MA_useUserDictionary",
		NH_MA_useUserDictionary, "iI(S)", 0);

	expf.PWPRegisterAyaoriFunc2(
		instance,
		"NH_MA_parse",
		NH_MA_parse, "iI(SS]])", 0);

	expf.PWPRegisterAyaoriFunc2(
		instance,
		"NH_MA_getLastError",
		NH_MA_getLastError, "iS()", 0);


	instance->pdata = pdata;
	return 1;
}


/**
 * vOCCX^Xj
 */
PWPEXPORT void PWPCALL
PWCDestroyInstance(TPWPInstance *instance)
{
	if(instance->pdata)
	{
		PData *pdata = (PData *)(instance->pdata);

		if(pdata->lastError)
		{
			free(pdata->lastError);
		}

		free(instance->pdata);
		instance->pdata = NULL;
	}
}


/**
 * vOCj
 */
PWPEXPORT void PWPCALL
PWCDestroy()
{
	if(mf)
	{
		MeCabDLL_unload(mf);
		mf = NULL;
	}
}


/*
 * --------------------------------------------------------------------------
 * CxgR[obN
 * --------------------------------------------------------------------------
 */
/*
PWPEXPORT void PWPCALL
PWCStartUp(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCChangedStartUp(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCShutDown(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCMouseUpL(TPWPInstance *instance, int cx, int cy) {}

PWPEXPORT void PWPCALL
PWCMouseUpR(TPWPInstance *instance, int cx, int cy) {}

PWPEXPORT void PWPCALL
PWCSilentModeOn(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCSilentModeOff(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCScriptUpdated(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCScriptUpdating(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCMouseEnter(TPWPInstance *instance, const char *name) {}

PWPEXPORT void PWPCALL
PWCMouseLeave(TPWPInstance *instance, const char *name) {}

PWPEXPORT void PWPCALL
PWCDragAndDrop(TPWPInstance *instance, DWORD hdrop) {}

PWPEXPORT void PWPCALL
PWCPersonaMinimize(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCPersonaRestore(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCPersonaDragged(TPWPInstance *instance) {}

PWPEXPORT void PWPCALL
PWCPersonaDropped(TPWPInstance *instance) {}

// Chararina 3.10 ȍ~
PWPEXPORT void PWPCALL
PWCMouseUpL2(TPWPInstance *instance, const char *windowname, int cx, int cy) {}

PWPEXPORT void PWPCALL
PWCMouseUpR2(TPWPInstance *instance, const char *windowname, nt cx, int cy) {}

PWPEXPORT void PWPCALL
PWCMouseEnter2(TPWPInstance *instance, const char *windowname, const char *name) {}

PWPEXPORT void PWPCALL
PWCMouseLeave2(TPWPInstance *instance, const char *windowname, const char *name) {}

PWPEXPORT void PWPCALL
PWCDragged2(TPWPInstance *instance, const char *windowname) {}

PWPEXPORT void PWPCALL
PWCDropped2(TPWPInstance *instance, const char *windowname) {}
*/


#ifdef __cplusplus
} /* extern "C" */
#endif
